/*                                                                      
 *  C-SaCzech
 *
 *  Copyright (c) 1996-2000 Jaromir Dolecek <dolecek@ics.muni.cz>           
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ----
 * Note:
 *   If the conditions of GPL are unacceptable for you, contact
 *   author to adjust something more suitable for your needs.
 */

/* $Id: cstools.h,v 1.34 2000/01/24 13:18:01 dolecek Exp $ */

#ifndef _CSA_CSTOOLS_H_
#define _CSA_CSTOOLS_H_

/* used to hold Unicode char */
typedef unsigned int cstools_unicode_t;

#define CSTOOLS_NUMFONTS	13
#define CSTOOLS_Unicode_base	128
typedef enum {
	CSTOOLS_ASCII	  = 0,
	CSTOOLS_ISOLatin1 = 1,
	CSTOOLS_ISOLatin2 = 2,
	CSTOOLS_CP1250	  = 3,
	CSTOOLS_Kam	  = 4,
	CSTOOLS_PCLatin2  = 5,
	CSTOOLS_KOI8_CS	  = 6,
	CSTOOLS_Mac	  = 7,
	CSTOOLS_CP850	  = 8,
	CSTOOLS_KOI8_R	  = 9,
	CSTOOLS_ISO8859_5 = 10,
	CSTOOLS_CP866	  = 11,
	CSTOOLS_CP1251	  = 12, /* CSTOOLS_MAX8BIT */

	CSTOOLS_Unicode	  = (CSTOOLS_Unicode_base + 0),
	CSTOOLS_UTF8	  = (CSTOOLS_Unicode_base + 1),

	CSTOOLS_MINCODE	  = 0,
	CSTOOLS_MAX8BIT	  = 12,
	CSTOOLS_UNKNOWN	  = -1,
	CSTOOLS_BINARY	  = -2
} cstools_t;

typedef enum {
	CSTOOLS_BOTHUNI, CSTOOLS_FROMUNI, CSTOOLS_TOUNI, CSTOOLS_NOUNI
} cstools_typeconv_t;

/* holds infomations used for recoding; initialized by cstools_setup() */
typedef struct {
        cstools_t source;		/* source encoding */
        cstools_t target;		/* target encoding */
	cstools_typeconv_t typeconv;	/* type of conversion */
        cstools_unicode_t map[256];	/* recoding map */
} cstools_cstocs_t;

#ifdef CSACEK
#include "csacek.h"
#else  /* !CSACEK */
#include <stdio.h>	/* for NULL, yup ;) */
#include <string.h>
#endif /* CSACEK */

#ifndef __P
#ifdef __STDC__
#  define	__P(x)	x
#else
#  define	__P(x)	()
#endif
#endif /* notdef __P */

#ifndef CSACEK
#define CSA_UPPER(c)	((c) >= 'a' && (c) <= 'z' ? (c) & 0xdf : (c))
#define CSA_ISHEXA(x) ( ((x) >= '0' && (x) <= '9') || ( (CSA_UPPER(x) >= 'A') && (CSA_UPPER(x) <= 'F') ) )
#define CSA_HEX2DEC(x) ((x) >= 'A' ? (((x) & 0xdf) - 'A')+10 : ((x) - '0'))
#endif

#define CSTOOLS_ISUNICODE(x) ((x) == CSTOOLS_Unicode || (x) == CSTOOLS_UTF8)
#define CSTOOLS_ISCYRILLIC(x) ((x) == CSTOOLS_KOI8_R || (x) == CSTOOLS_CP866 \
	|| (x) == CSTOOLS_CP1251)

#define CSTOOLS_NEED_SETUP(mp, s, t) (mp.source != s || mp.target != t)

/* constants used as second parameter to cstools_name() */
#define CSTOOLS_TRUENAME	0
#define CSTOOLS_CODENAME	1
#define CSTOOLS_ABBRNAME	2
#define CSTOOLS_MIMENAME	3
#define CSTOOLS_LAMPANAME	4

/* I'd rather use this - toupper() could not be available */
#define CSTOOLS_UPPER(x) (((x) >= 97 && (x) <= 122) ? (x)-32 : (x))

/* functions */

#ifdef __cplusplus
extern "C" {
#endif

/* cstools.c */
cstools_t cstools_whichcode __P((const char *, size_t));
int cstools_recode __P((cstools_cstocs_t *, const char *, char *, int));
int cstools_init __P((cstools_cstocs_t *mp, cstools_t from, cstools_t to));
cstools_t cstools_guess_charset __P((const char *, size_t));

const char *cstools_name __P((cstools_t code, int which));

cstools_t cstools_index2code __P((int));
int cstools_code2index __P((cstools_t));

#ifdef __cplusplus
}
#endif

#endif /* _CSA_CSTOOLS_H_ */
