#line 2 "debug.c"
/*                                                                      
 *  C-SaCzech
 *
 *  Copyright (c) 1996-2000 Jaromir Dolecek <dolecek@ics.muni.cz>           
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ----
 * Note:
 *   If the conditions of GPL are unacceptable for you, contact
 *   author to adjust something more suitable for your needs.
 */

/* $Id: debug.c,v 1.36 2000/01/24 13:18:02 dolecek Exp $ */

#include "csacek.h"

#ifdef CSA_DEBUG

static size_t debug_no=0;		/* debug log number */

/* initialize debugfd and a few other needed things */
FILE *
csa_debug_start()
{
   const char *prefix, *suffix;
   char buf[2048];
   FILE *dd;
   size_t debug_num;

   prefix = getenv("TEMP");
   if (!prefix) {
	prefix = getenv("TMP");
   	if (!prefix) {
#ifdef __MSWIN__
   		prefix = "C:\\";
#else
   		prefix = "/tmp";
#endif
	}
   }

#ifdef __MSWIN__
   suffix = ".csa";
#else
   suffix = "";
#endif

   csa_acquire_mutex(csa_debuglog_mutex);
   debug_num = debug_no++;
   csa_release_mutex(csa_debuglog_mutex);
   
   sprintf(buf, "%s/csacek_debug_%d_%d%s", prefix, (int) getpid(), debug_num,
		suffix);

   dd = fopen(buf, "wb");
   if (dd) {
   	/* do NOT buffer debug log  - program can crash any time */
#ifndef SETVBUF_REVERSED
	setvbuf(dd, NULL, _IONBF, 0 ); /* "normal" form */
#else
	setvbuf(dd, _IONBF, NULL, 0 ); /* reversed form */
#endif

	csa_debug(dd, "--- BEGIN DEBUG LOG ---");
   }

   return dd;
}
	
/*
 * write a message to debug log
 */
void 
#ifdef __ANSI_C__
csa_debug(FILE *dbg, const char *fmt, ...)
#else
csa_debug(dbg, fmt /*, ... */)
  FILE *dbg;
  const char *fmt;
#endif /* __ANSI_C__ */
{
  va_list arg;

  fputs("DEBUG: ", dbg);
  va_start(arg, fmt);
#ifndef __ANSI_C__
  fmt = va_arg(arg, char *);
#endif
  vfprintf(dbg, fmt, arg);
  va_end(arg);
  fputs("\n", dbg);
}

void 
csa_debug_end(dbg)
  FILE *dbg;
{
	csa_debug(dbg, "--- END DEBUG LOG ---");
	fclose(dbg);
}

static void x_debug_end __P((void *));
static void x_debug_child_end __P((void *));

static void
x_debug_end(dbg)
  void * dbg;
{
	csa_debug_end((FILE *)dbg);
}

static void
x_debug_child_end(dbg)
  void *dbg;
{
	/*
	 * If the dbg stream would be buffered, this would
	 * probably cause that some data would be duplicated
	 * on child cleanup in debug log (the underlying buffer
	 * would be flushed in parent and in child as well),
	 * but since it's NOT buffered, there is no problem.
	 */
	fclose((FILE *)dbg);
}

/* schedules closing of the debug log file upon the pool destroy */
void csa_debug_register_cleanup(wpool, dbg)
  struct pool *wpool;
  FILE *dbg;
{
	ap_register_cleanup(wpool, (void *)dbg, x_debug_end,
		x_debug_child_end);
}
#endif /* CSA_DEBUG */
