#line 2 "info.c"
/*                                                                      
 *  C-SaCzech
 *
 *  Copyright (c) 1996-2000 Jaromir Dolecek <dolecek@ics.muni.cz>           
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ----
 * Note:
 *   If the conditions of GPL are unacceptable for you, contact
 *   author to adjust something more suitable for your needs.
 */

/* $Id: info.c,v 1.88 2000/01/24 13:18:06 dolecek Exp $ */

#include "csacek.h"
#include "csa_version.h"

#ifdef CSA_WANT_INFO
static const char *x_htmlify __P((const char *, char *));

/*
 * substitutes all occurences of > with &gt; and < with &lt;
 */
static const char *
x_htmlify(what, buff)
  const char *what;
  char *buff;
{
	const char *orig_buff=buff;
	size_t len;

	while(*what) {
		len = strcspn(what, "<>");
		if (len > 0) {
			strncpy(buff, what, len);
			what += len;
			buff += len;
		}

		if (*what == '>') strncpy(buff, "&gt;", 4);
		else if (*what == '<') strncpy(buff, "&lt;", 4);
		else
			continue;
		what++;
		buff += 4;
	}
	*buff = '\0';
	return orig_buff;
}

/*
 * generate report page about compiled-in defaults 
 */
int 
csa_info(p, cfg)
  csa_params_t *p;
  const csa_conf_t *cfg;
{
	char buf[2048], buf2[2048];
	size_t len;

#ifdef CSA_DEBUG
	csa_debug(p->dbg, "csa_info: called");
#endif

	csa_setheaderout(p, "Status", "200 OK", 0);
	csa_setheaderout(p, "Content-Type", "text/html", 0);

        len = sprintf(buf, "<HTML>\n<HEAD>\n<TITLE>CSaCzech %s report</TITLE>\n</HEAD>\n<BODY BGCOLOR=WHITE>\n",
                        CSA_VERSION);
	csa_add_output(p, buf, len, 0);

	len = sprintf(buf, "<H3><A HREF=\"http://www.csacek.cz/\">CSaCzech %s</A> (c) 1996-2000 <a href=\"http://www.ics.muni.cz/~dolecek/\">Jaromir Dolecek</a></H3>\n", CSA_VERSION );
	csa_add_output(p, buf, len, 0);

# ifdef CSA_DEBUG
	csa_add_output(p, "This is <STRONG>DEBUG</STRONG> version.<BR>\n",
		0, CSA_OUT_STR);
#endif /* CSA_DEBUG */
	len = sprintf(buf,"This version has been compiled on <TT>%s</TT> (%s) as %s.\n",
		CSA_SYSTEM_TYPE, CSA_OS_TYPE, CSA_MUTACE_NAME);
	csa_add_output(p, buf, len, 0);

	/* CSacek configuration */
	csa_add_output(p, "<H4>Configuration:</H4>\n", 0, CSA_OUT_STR);

	len = sprintf(buf, "Compression: <TT>%s</TT><BR>\n",
		(cfg->compress) ? "enabled" : "disabled");
	csa_add_output(p, buf, len, 0);

	len = sprintf(buf, "Default source encoding: <TT>%s</TT><BR>\n", 
		cfg->DefaultCharset);
	csa_add_output(p, buf, len, 0);

	len = sprintf(buf, "TemplateDir: <TT>%s</TT><BR>\n",
		cfg->TemplateDir);
	csa_add_output(p, buf, len, 0);

	len = sprintf(buf, "IgnorePrefix: <TT>%s</TT><BR>\n",
		p->ignoreprefix->value);
	csa_add_output(p, buf, len, 0);	

	len = sprintf(buf, "Default part: <TT>%s</TT><BR>\n",
		cfg->DefaultPartname);
	csa_add_output(p, buf, len, 0);

	len = sprintf(buf, "Default BAR settings: <TT>%s</TT><BR>\n",
		(cfg->BarDef) ?  x_htmlify(cfg->BarDef, buf2)
			: "(not specified)");
	csa_add_output(p, buf, len, 0);

	len = sprintf(buf, "URL rewriting: <TT>%s</TT><BR>\n",
		(cfg->changeurl) ? "on" : "off");
	csa_add_output(p, buf, len, 0);

	len = sprintf(buf,
		"Recoding client data to <TT>%s</TT>: <TT>%s</TT><BR>\n",
		cfg->DefaultCharset,
		(cfg->recodeinput) ? "on" : "off");
	csa_add_output(p, buf, len, 0);

	len = sprintf(buf,
		"Implicit work: <TT>%s</TT><BR>\n",
		(cfg->implicitWork ? "on" : "off"));
	csa_add_output(p, buf, len, 0);

	/* list of CSacek servers */
	csa_add_output(p, "Registered CSacek servers:\n",
		0, CSA_OUT_STR);
	csa_write_servers(p);

	/* BAR settings */
	csa_add_output(p, "<H4>BAR settings:</H4>\n", 0, CSA_OUT_STR);

#ifdef CSA_ALWAYSADDBAR
	csa_add_output(p, "<TT>BAR</TT> is automagically added on the end of any HTML page.<BR>\n", 0, CSA_OUT_STR);
#endif

	len = sprintf(buf,
		"Names of codes in bar are %s by&nbsp;default.<BR>\n",
#ifndef CSA_LONGNAMEBAR
		"long"
#else
		"short"
#endif
	);
	csa_add_output(p, buf, len, 0);
	len = sprintf(buf,
		"HTML code at head of the czech bar: <TT>%s</TT><BR>", 
		x_htmlify(CSA_HEADBARTEXT, buf2));
	csa_add_output(p, buf, len, 0);
	len = sprintf(buf,
		"HTML code at head of the english bar: <TT>%s</TT><BR>\n",
		x_htmlify(CSA_EHEADBARTEXT, buf2));
	csa_add_output(p, buf, len, 0);
	len = sprintf(buf,
		"HTML code at end of czech bar is: <TT>%s</TT><BR>\n",
		x_htmlify(CSA_TAILBARTEXT, buf2));
	csa_add_output(p, buf, len, 0);
	len = sprintf(buf,"HTML code at end of english bar: <TT>%s</TT><BR>\n",
		x_htmlify(CSA_ETAILBARTEXT, buf2));
	csa_add_output(p, buf, len, 0);
	len = sprintf(buf,
		"Link to whichcode is%s added to bar by default.<BR>\n",
#  ifdef CSA_ADDWHICHCODE
		""
#  else
		"n't"
#  endif
	);
	csa_add_output(p, buf, len, 0);
	len = sprintf(buf,
		"Name of whichcode option in a bar is '%s' by default.<BR>\n",
#  ifdef CSA_BARWHICHCODENAME
		CSA_BARWHICHCODENAME
#  else
		"[none]"
#  endif
	);
	csa_add_output(p, buf, len, 0);

	/* other settings */
	csa_add_output(p, "<P><H4>Other settings:</H4>\n", 0, CSA_OUT_STR);

#ifdef CSA_MUTACE_XCGI
	len = sprintf(buf, "Documents with following suffixes will be coded by method \"file\": <EM>%s</EM>.<BR>\n", CSA_CONVERT_SUFFIXES);
	csa_add_output(p, buf, len, 0);

	len = sprintf(buf, "Suffixes <EM>\"%s\"</EM> are ignored in name of CSacek script and won't ever be&nbsp;a partname.<BR>\n", CSA_IGNORE_SUFFIXES );
	csa_add_output(p, buf, len, 0);

	len = sprintf(buf, "Getmethod is '%s'.<BR>\n", 
# if CSA_METHOD_HTTP
		"http"
# elif CSA_METHOD_FILE
		"file"
# elif CSA_METHOD_GUESS
		"guess"
# else
		"unknown"
# endif
	);
	csa_add_output(p, buf, len, 0);
#endif /* CSA_MUTACE_XCGI */

	len = sprintf(buf,
	"The character which marks C-SaCzech commands is '<TT>%c</TT>'.<BR>\n",
		CSA_CMD_TAG);
	csa_add_output(p, buf, len, 0);

	len = sprintf(buf, "Optional parameter 'charset' %s used in returned <TT>Content-Type</TT> header.<BR>\n",
# ifdef CSA_USE_CHARSET
		"IS"
# else
		"is NOT"
# endif
	);
	csa_add_output(p, buf, len, 0);

#ifdef CSA_MUTACE_XCGI
	len = sprintf( buf, "Security: %s%s.<BR>\n",
# ifdef CSA_WANT_SECURE
		"I DO care about it; I look for files '.htaccess' & ",
		"'.nsconfig' on path to accessed document."
# else
		"I do NOT care about this", ""
# endif
	);
	csa_add_output(p, buf, len, 0);
#endif /* CSA_MUTACE_(FAST)CGI */

#ifdef CSA_CANONIFY_DIRURL
	csa_add_output(p, "If requested URL leads to dir and doesn't end with <TT>/</TT>, I redirect client to&nbsp;the&nbsp;right one.<BR>\n", 0, CSA_OUT_STR);
#endif

	csa_add_output(p, "<TT>__LAMPACHARSET__</TT> is substituted by&nbsp;name of&nbsp;target encoding\nin&nbsp;form used by&nbsp;Lampa's Apache czech module.<BR>\n",
		0, CSA_OUT_STR);

#ifdef CSA_COMPAT_13X
	csa_add_output(p, "I try to be compatible with&nbsp;older (pre 1.3.3) versions of C-SaCzech;\nolder syntax of <TT>PART, DOMAIN</TT> and <TT>BAR</TT> would work.<BR>\n", 0, CSA_OUT_STR);
#endif

#ifdef CSA_WANT_BANNER
	csa_add_output(p, "I write short banner at the start of any recoded HTML page to mark that it has been recoded by CSacek.\n", 0, CSA_OUT_STR);
#endif

	csa_add_output(p, "</BODY>\n</HTML>", 0, CSA_OUT_STR);

	return CSA_INFO;
} /* csa_report */
#endif /* CSA_WANT_INFO */

void 
csa_version()
{
	printf("C-SaCzech %s %s\nCopyright (c) 1996-2000 Jaromir Dolecek <dolecek@ics.muni.cz>\n", CSA_VERSION,
#ifdef CSA_DEBUG
		"(DEBUG)"
#else
		""
#endif /* CSA_DEBUG */
		 );
	printf("	Built on: %s, %s\n", CSA_OS_TYPE, CSA_SYSTEM_TYPE);
}
